#include <TFT.h> 

// Definirea pinilor
#define cs 10
#define dc 9
#define rst 8
#define TFT_PURPLE 0xF8DBF8
#define TFT_BLUE 0x12A722
#define TFT_BLACK 0x000000
#define TFT_WHITE 0xFFFFFF
#define TFT_RED 0xFF0000
#define TFT_GREEN 0x00FF00

// instanta a librariei TFT ce asigura comunicarea la ecranul
TFT TFTscreen = TFT(cs, dc, rst);

// variabile necesare primei functionalitati
char alphabet[] = "ABCDEFGHIJLMNOPRSTUVWXYZ";
char alphabetMatrix[4][8] = {
  { 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H' },
  { 'I', 'J', 'L', 'M', 'N', 'O', 'P', 'R' },
  { 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z' },
};

// configurarea pinilor pentru joystick
const int VRx = A0;
const int VRy = A1;
const int SW = A2;
int cursorRow = 0;
int cursorCol = 0;
bool nameEntered = false;
String enteredName = "";
bool startPage = true;

// configurarea pinului pentru buzzer
int buzzerPin = A3;

// definirea pinilor pentru butoane
const int BUTTON_PIN_1 = 2;  // Pinul pentru primul buton
const int BUTTON_PIN_2 = 3;  // Pinul pentru al doilea buton
const int BUTTON_PIN_3 = 4;  // Pinul pentru al treilea buton
const int BUTTON_PIN_4 = 5;  // Pinul pentru al patrulea buton
const int BUTTON_PIN_5 = 6;  // Pinul pentru al cincilea buton
const int BUTTON_PIN_6 = 7;  // Pinul pentru al șaselea buton

// variabile necesare pentru desfasurarea jocului
bool gameStarted = false;
const int table_rows = 6;  // numărul de rânduri ale tablei
const int table_cols = 9;  // numărul de coloane ale tablei
// pozitia initiala a cursorului
int tableCursorRow = 3;
int tableCursorCol = 5;
// dimensiunea celulelor tablei
int cellSize = 15;  
int boardX = 10;    // coordonata x de start pentru tabla
int boardY = 20;    // coordonata y de start pentru tabla
int cellX;
int cellY;
int score;
int prevCursorRow = -1;
int prevCursorCol = -1;

// variabile pt debouncing in apasarea butoanelor
unsigned long lastMoveTime = 0;   
unsigned long moveInterval = 200;

// matricele de joc 
int stateBoard[table_rows][table_cols] = { 0 };
int board[table_rows][table_cols] = { 0 };


void setup() {
  TFTscreen.begin();
  clean_background();

  // initializare joystick-ul
  pinMode(SW, INPUT);
  digitalWrite(SW, HIGH);

  // initializare buzzer-ul
  pinMode(buzzerPin, OUTPUT);
  digitalWrite(buzzerPin, HIGH);

  // initializare butoane
  pinMode(BUTTON_PIN_1, INPUT_PULLUP);
  pinMode(BUTTON_PIN_2, INPUT_PULLUP);
  pinMode(BUTTON_PIN_3, INPUT_PULLUP);
  pinMode(BUTTON_PIN_4, INPUT_PULLUP);
  pinMode(BUTTON_PIN_5, INPUT_PULLUP);
  pinMode(BUTTON_PIN_6, INPUT_PULLUP);

  // desenare initiala tabla
  drawAlphabetScreen();
  drawBorder(enteredName);
  randomSeed(analogRead(3));
}

void loop() {

  if (startPage == true) {
    // citirea stării butoanelor
    int button1Pressed = digitalRead(BUTTON_PIN_1);
    int button2Pressed = digitalRead(BUTTON_PIN_2);

    // apasarea butonului 1, duce la pagina jocului propriu-zis
    if (button1Pressed == LOW) {
      positiveSound();
      startPage = false;
    }

    // citiri joystick
    int xValue = analogRead(VRx);
    int yValue = analogRead(VRy);
    bool buttonPressed = analogRead(SW) < 512;  
    // timp de așteptare pentru debounce
    delay(200);

    // actualizarea poziția cursorului în funcție de intrările joystick-ului
    if (xValue < 512 - 200) {
      // la dreapta
      cursorCol = min(7, cursorCol + 1);  
    } else if (xValue > 512 + 200) {
      // la stânga
      cursorCol = max(0, cursorCol - 1);  
    }
    if (yValue < 512 - 200) {
      // în jos
      cursorRow = min(2, cursorRow + 1); 
    } else if (yValue > 512 + 200) {
      // în sus
      cursorRow = max(0, cursorRow - 1);  
    }

    // adaugarea unei litere din nume
    if (buttonPressed) {
      oneSound();
      char selectedLetter = alphabetMatrix[cursorRow][cursorCol];
      enteredName += selectedLetter;  
    }

    // ștergerea unei litere din nume
    if (button2Pressed == LOW && enteredName.length() > 0) {
      oneSound();
      clean_background();
      enteredName.remove(enteredName.length() - 1);
    }

    // desenează matricea de litere cu litera curentă evidențiată
    drawAlphabetScreen();
    // deseneaza chenarul cu numele
    drawBorder(enteredName); 
  } else {


    // JOCUL PROPRIU-ZIS
    // citirea butoanelor
    int button1Pressed = digitalRead(BUTTON_PIN_1);
    int button2Pressed = digitalRead(BUTTON_PIN_2);
    int button3Pressed = digitalRead(BUTTON_PIN_3);
    int button4Pressed = digitalRead(BUTTON_PIN_4);
    int button5Pressed = digitalRead(BUTTON_PIN_5);
    int button6Pressed = digitalRead(BUTTON_PIN_6);

    // desenare elementelor de baza pt ecranul jocului
    if (!gameStarted) {
      score = 0;
      initializeMinesweeperBoard();
      drawScore(enteredName);
      gameStarted = true;
      prevCursorRow = tableCursorRow;
      prevCursorCol = tableCursorCol;
      redrawCursor();
    }
    
    if (button5Pressed == LOW && millis() - lastMoveTime >= moveInterval) {
      // Deplasare cursor spre STG
      oneSound();
      tableCursorCol = max(tableCursorCol - 1, 0);
      lastMoveTime = millis();
      redrawCursor();

    } else if (button4Pressed == LOW && millis() - lastMoveTime >= moveInterval) {
      // Deplasare cursor spre DR
      oneSound();
      tableCursorCol = min(tableCursorCol + 1, 8);
      lastMoveTime = millis();
      redrawCursor();

    } else if (button3Pressed == LOW && millis() - lastMoveTime >= moveInterval) {
      // Deplasare cursor în SUS
      oneSound();
      tableCursorRow = max(tableCursorRow - 1, 0);
      lastMoveTime = millis();
      redrawCursor();

    } else if (button6Pressed == LOW && millis() - lastMoveTime >= moveInterval) {
      // Deplasare cursor în JOS
      oneSound();
      tableCursorRow = min(tableCursorRow + 1, 5);
      lastMoveTime = millis();
      redrawCursor();
    }

    if (button1Pressed == LOW) {
      // Descoperirea unui patrat
      oneSound();
      revealCell(tableCursorRow, tableCursorCol);
      drawScore(enteredName);
      delay(100);

    } else if (button2Pressed == LOW) {
      // Marcarea ca flag
      oneSound();
      toggleFlag(tableCursorRow, tableCursorCol);
      drawFlag(tableCursorRow, tableCursorCol);
      delay(100); 
    }

    // Verific dacă jocul este câștigat
    if (checkWin()) {
      displayWinMessage();
      positiveSound();
      while (true) {
        // Blochez execuția buclei pentru a nu permite alte acțiuni după câștig
      }
    }
    
  }
}

// functie care sterge ultima litera introdusa pt ecranului de inceput
void clean_background() {
  TFTscreen.fillScreen(TFT_PURPLE);
  TFTscreen.setTextSize(2);
  TFTscreen.stroke(0, 0, 0);
  TFTscreen.text(" MINESWEEPER", 5, 5);
  TFTscreen.setTextSize(1);
  TFTscreen.text("Please enter your name:", 10, 23);
}

// functie care deseneaza tastatura cu litere
void drawAlphabetScreen() {
  int startX = 5;          
  int startY = 40;         
  int letterSpacing = 20;  
  int rowHeight = 20;      
  int numRows = 4;
  int numCols = 8;

  for (int row = 0; row < numRows; row++) {
    for (int col = 0; col < numCols; col++) {
      int x = startX + col * letterSpacing;
      int y = startY + row * rowHeight;

      if (row == cursorRow && col == cursorCol) {
        // litera selectata este de culoare rosie
        TFTscreen.stroke(0, 0, 255);  
      } else {
        // culoarea celorlalte litere este alabastru
        TFTscreen.stroke(255, 0, 0); 
      }
      TFTscreen.setTextSize(2);
      char letter = alphabetMatrix[row][col];
      // afisez litera curenta
      TFTscreen.text(String(letter).c_str(), x, y);
    }
  }
}

// functie care deseneaza chenarul in care este intodus numele
void drawBorder(String enteredName) {
  int chenarX = 10;  
  int chenarY = 100;     
  int chenarWidth = 140;
  int chenarHeight = 20;

  // Desenez chenarul negru
  TFTscreen.stroke(0, 0, 0); 
  TFTscreen.noFill();
  TFTscreen.rect(chenarX, chenarY, chenarWidth, chenarHeight);

  // Afisez numele
  TFTscreen.setTextSize(1);                                
  TFTscreen.stroke(0, 0, 0);                     
  TFTscreen.text(enteredName.c_str(), chenarX + 5, chenarY + 5);
}

// functie care emite un sunet pozitiv
void positiveSound() {
  digitalWrite(buzzerPin, LOW);
  delay(50);
  digitalWrite(buzzerPin, HIGH);
  delay(300);
  digitalWrite(buzzerPin, LOW);
  delay(50);
  digitalWrite(buzzerPin, HIGH);
  delay(300);
  digitalWrite(buzzerPin, LOW);
  delay(50);
  digitalWrite(buzzerPin, HIGH);
}

// functie care emite un sunet negativ
void negativeSound() {
  digitalWrite(buzzerPin, LOW);
  delay(50);
  digitalWrite(buzzerPin, HIGH);
  delay(1000);
  digitalWrite(buzzerPin, LOW);
  delay(50);
  digitalWrite(buzzerPin, HIGH);
}

// functie care emite un singur sunet
void oneSound() {
  digitalWrite(buzzerPin, LOW);
  delay(50);
  digitalWrite(buzzerPin, HIGH);
}

// functie ce pregateste jocul propriu-zis
void initializeMinesweeperBoard() {
  TFTscreen.fillScreen(TFT_BLUE);
  // initializarea matricilor
  for (int i = 0; i < table_rows; i++) {
    for (int j = 0; j < table_cols; j++) {
      board[i][j] = 0;
      stateBoard[i][j] = 0;
    }
  }

  // apelul functiilor pt plasarea bombelor
  placeMines(board, table_rows, table_cols);
  // si pentru calculul valorilor
  calculateNeighborMines(board, table_rows, table_cols);

  // desenarea celulelor tablei
  for (int i = 0; i < table_rows; i++) {
    for (int j = 0; j < table_cols; j++) {
      int cellX = boardX + j * cellSize;                 
      int cellY = boardY + i * cellSize;                 
      TFTscreen.stroke(0, 0, 0);                         
      TFTscreen.noFill();                                
      TFTscreen.rect(cellX, cellY, cellSize, cellSize);  
    }
  }
}

// functie pentru mutarea cursorului in jocul propriu-zis
void redrawCursor() {
  // sterg pozitia veche a cursorului
  int prevCellX = boardX + prevCursorCol * cellSize;
  int prevCellY = boardY + prevCursorRow * cellSize;

  if (stateBoard[prevCursorRow][prevCursorCol] == 1) {
    // daca celula anterioare a fost o celula descoperita, o redesenez folosind culoarea specifica si datele specifice
    TFTscreen.fill(200, 200, 200);
    TFTscreen.stroke(0, 0, 0);
    TFTscreen.rect(prevCellX, prevCellY, cellSize, cellSize);
    int numNeighbors = board[prevCursorRow][prevCursorCol];
    if (numNeighbors > 0) {
      TFTscreen.text(String(numNeighbors).c_str(), prevCellX + cellSize / 2 - 3, prevCellY + cellSize / 2 - 3);
    }
  } else {
    // altfel nimic deosebit
    TFTscreen.noFill(); 
    TFTscreen.stroke(0, 0, 0);  
    TFTscreen.rect(prevCellX, prevCellY, cellSize, cellSize);
  }

  // desenez pozitia curenta a cursorului
  int cellX = boardX + tableCursorCol * cellSize;
  int cellY = boardY + tableCursorRow * cellSize;
  TFTscreen.stroke(0, 0, 255);  // chenar de culoare rosie
  TFTscreen.noFill();  
  TFTscreen.rect(cellX, cellY, cellSize, cellSize);

  // Updatez pozitia aneterioara
  prevCursorRow = tableCursorRow;
  prevCursorCol = tableCursorCol;
}

// functie pentru desenarea continua score-ului
void drawScore(String enteredName) {
  TFTscreen.fillRect(10, 10, 100, 10, TFT_BLUE);
  TFTscreen.setTextSize(1);   
  TFTscreen.stroke(0, 0, 0);  
  enteredName += "'s score: ";
  enteredName += String(score);
  TFTscreen.text(enteredName.c_str(), 10, 10);
}

// functie ce plaseaza aleatoriu bombele
void placeMines(int board[][9], int rows, int cols) {
  int numMines = 7; 
  int placedMines = 0;

  while (placedMines < numMines) {
    int randRow = random(rows);  
    int randCol = random(cols);  

    if (board[randRow][randCol] != 9) {
      // cifra 9 in matrice reprezinta existenta unei mine
      //o celula nedescoperita are valoare 0, iar una descoperita are valoarea 1
      board[randRow][randCol] = 9;  
      placedMines++;                
    }
  }
}

// functie ce calculeaza valorile celulelor invecinate cu bombele
void calculateNeighborMines(int board[][9], int rows, int cols) {
  for (int i = 0; i < rows; i++) {
    for (int j = 0; j < cols; j++) {
      // cazul in care celula nu contine o bomba
      if (board[i][j] != 9) {
        int neighborMines = 0;

        // Parcurg celulele vecine (maxim 8)
        for (int dx = -1; dx <= 1; dx++) {
          for (int dy = -1; dy <= 1; dy++) {
            int ni = i + dx;
            int nj = j + dy;

            // Verific dacă celula vecină este validă și conține o bomba
            if (ni >= 0 && ni < rows && nj >= 0 && nj < cols && board[ni][nj] == 9) {
              neighborMines++;
            }
          }
        }

        // Actualizez nr de bombe vecine pentru celula curentă
        board[i][j] = neighborMines;
      }
    }
  }
}


// functie care marcheaza existanta unui flag in matrice
void toggleFlag(int row, int col) {
  if (stateBoard[row][col] == 0) {
    stateBoard[row][col] = 2; 
  } 
}

// functie care deseneaza flag-ul
void drawFlag(int row, int col) {
  int cellX = boardX + col * cellSize;
  int cellY = boardY + row * cellSize;
  TFTscreen.setTextSize(1);
  TFTscreen.fill(200, 200, 200);
  TFTscreen.rect(cellX, cellY, cellSize, cellSize);
  // flag-ul este albastru
  TFTscreen.stroke(255, 0, 0);
  TFTscreen.text("F", cellX + cellSize / 2 - 3, cellY + cellSize / 2 - 3);
}

// functie pentru descoperirea celulelor
void revealCell(int row, int col) {
  // daca celula nu este descoperita
  if (stateBoard[row][col] == 0) {  
    // o marchez ca fiind descoperita
    stateBoard[row][col] = 1;       

    int cellX = boardX + col * cellSize;
    int cellY = boardY + row * cellSize;
    TFTscreen.setTextSize(1);

    // descoperirea unei bomabe duce la sfarsitul jocului
    if (board[row][col] == 9) {
      TFTscreen.fill(200, 200, 200);
      TFTscreen.rect(cellX, cellY, cellSize, cellSize);
      TFTscreen.stroke(0, 0, 255);
      TFTscreen.text("X", cellX + cellSize / 2 - 3, cellY + cellSize / 2 - 3);
      delay(4000);
      displayGameOverMessage();
      negativeSound();
      while (true) {
        // Blochez execuția buclei pentru a nu permite alte acțiuni 
      }

    } else {
      int numNeighbors = board[row][col];
      // daca se descopera o celula cu o valoarea exacta, este afisata
      if (numNeighbors > 0) {
        // creste valoarea score-ului
        score += 5 * numNeighbors;
        TFTscreen.fill(200, 200, 200);
        TFTscreen.rect(cellX, cellY, cellSize, cellSize);
        TFTscreen.stroke(0, 0, 0);
        TFTscreen.text(String(numNeighbors).c_str(), cellX + cellSize / 2 - 3, cellY + cellSize / 2 - 3);
      } else {
        // daca celula nu are valoare, se face descoperire in masa
        score += 20;
        TFTscreen.stroke(0, 0, 0);
        // celulele descoperite sunt marcate cu o alta culoare
        TFTscreen.fill(200, 200, 200);  
        TFTscreen.rect(cellX, cellY, cellSize, cellSize);
        revealAdjacentCells(row, col);
      }
    }
  }
}

// functie auxiliara pentru descoperirea in masa
void revealAdjacentCells(int row, int col) {
  for (int dx = -1; dx <= 1; dx++) {
    for (int dy = -1; dy <= 1; dy++) {
      int ni = row + dx;
      int nj = col + dy;
      if (ni >= 0 && ni < table_rows && nj >= 0 && nj < table_cols && stateBoard[ni][nj] == 0) {
        revealCell(ni, nj);
      }
    }
  }
}

// functie care verifica daca jocul a fost castigat
bool checkWin() {
  for (int i = 0; i < table_rows; i++) {
    for (int j = 0; j < table_cols; j++) {
      if (board[i][j] == 9 && stateBoard[i][j] != 2) {
        // Dacă există o mină care nu este marcată cu un steag, nu s-a castigat
        return false;
      }
      if (board[i][j] != 9 && stateBoard[i][j] != 1) {
        // Dacă există o celulă fără bomba care nu este descoperită, nu s-a castigat
        return false;
      }
    }
  }
  return true;  
}

// functie pentru ecranul afisat in caz de castig
void displayWinMessage() {
  TFTscreen.background(0, 255, 0);
  TFTscreen.stroke(0, 0, 0);
  TFTscreen.setTextSize(2);
  int textX = (TFTscreen.width() - (strlen("You win!") * 12)) / 2; 
  int textY = (TFTscreen.height() - 16) / 2; 
  TFTscreen.text("You Win!", textX, textY - 50);
  TFTscreen.text("Your score:", textX - 20, textY - 30);
  String s = String(score);
  TFTscreen.text(s.c_str(), textX + 40, textY - 5);

  // Coordonatele și dimensiunile pentru fața zâmbitoare
  int faceCenterX = TFTscreen.width() / 2;
  int faceCenterY = textY + 40; 
  int faceRadius = 20;

  // Desenează cercul pentru față
  TFTscreen.fill(0, 255, 255); // fata are culoarea galbena
  TFTscreen.stroke(0, 0, 0);
  TFTscreen.circle(faceCenterX, faceCenterY, faceRadius);

  // Desenez ochii
  int eyeRadius = 3;
  int eyeOffsetX = 7;
  int eyeOffsetY = -5;
  TFTscreen.fill(0, 0, 0); 
  TFTscreen.circle(faceCenterX - eyeOffsetX, faceCenterY + eyeOffsetY - 5, eyeRadius); 
  TFTscreen.circle(faceCenterX + eyeOffsetX, faceCenterY + eyeOffsetY - 5, eyeRadius);

  // Desenez gura zâmbitoare
  int mouthWidth = 25;
  int mouthHeight = 20; 

  TFTscreen.line(faceCenterX - mouthWidth / 2, faceCenterY + mouthHeight / 4, faceCenterX - mouthWidth / 4, faceCenterY + mouthHeight / 2);
  TFTscreen.line(faceCenterX - mouthWidth / 4, faceCenterY + mouthHeight / 2,   faceCenterX + mouthWidth / 4, faceCenterY + mouthHeight / 2);
  TFTscreen.line(faceCenterX + mouthWidth / 4, faceCenterY + mouthHeight / 2, faceCenterX + mouthWidth / 2, faceCenterY + mouthHeight / 4);
}

// functie pentru ecranul afisat in caz de esec
void displayGameOverMessage() {
  TFTscreen.background(0, 0, 255);
  TFTscreen.stroke(0, 0, 0);
  TFTscreen.setTextSize(2);
  int textX = (TFTscreen.width() - (strlen("Game Over") * 12)) / 2; 
  int textY = (TFTscreen.height() - 20) / 2; 
  TFTscreen.text("Game Over", textX, textY - 50);

  TFTscreen.text("Your score:", textX - 10, textY - 30);
  String s = String(score);
  TFTscreen.text(s.c_str(), textX + 50, textY - 5);

  // Coordonatele și dimensiunile pentru fața trista
  int faceCenterX = TFTscreen.width() / 2;
  int faceCenterY = textY + 40;
  int faceRadius = 20;

  // Desenez cercul pentru fața galbena
  TFTscreen.fill(0, 255, 255); 
  TFTscreen.stroke(0, 0, 0);   
  TFTscreen.circle(faceCenterX, faceCenterY, faceRadius);

  // Desenez ochii
  int eyeRadius = 3;
  int eyeOffsetX = 7;
  int eyeOffsetY = -5;
  TFTscreen.fill(0, 0, 0);
  TFTscreen.circle(faceCenterX - eyeOffsetX, faceCenterY + eyeOffsetY - 5, eyeRadius); 
  TFTscreen.circle(faceCenterX + eyeOffsetX, faceCenterY + eyeOffsetY - 5, eyeRadius);

  // Desenez gura trista
  int mouthWidth = 25; 
  int mouthHeight = 20; 

  TFTscreen.line(faceCenterX - mouthWidth / 2, faceCenterY + mouthHeight / 2, faceCenterX - mouthWidth / 4, faceCenterY + mouthHeight / 4);
  TFTscreen.line(faceCenterX - mouthWidth / 4, faceCenterY + mouthHeight / 4, faceCenterX + mouthWidth / 4, faceCenterY + mouthHeight / 4);
  TFTscreen.line(faceCenterX + mouthWidth / 4, faceCenterY + mouthHeight / 4, faceCenterX + mouthWidth / 2, faceCenterY + mouthHeight / 2);
}
